from pydantic import ValidationError

from ..utils import pop_keywords
from .base import BasePlugin, Context
from .page import PAGES


class FlaskPlugin(BasePlugin):
    def find_routes(self):
        for rule in self.app.url_map.iter_rules():
            if any(str(rule).startswith(path) for path in (
                f'/{self.config.PATH}', '/static'
            )):
                continue
            yield rule

    def bypass(self, func, method):
        if method in ['HEAD', 'OPTIONS']:
            return True
        return False

    def parse_func(self, route):
        func = self.app.view_functions[route.endpoint]
        for method in route.methods:
            yield method, func

    def parse_path(self, route):
        from werkzeug.routing import parse_rule, parse_converter_args

        subs = []
        parameters = []

        for converter, arguments, variable in parse_rule(str(route)):
            if converter is None:
                subs.append(variable)
                continue
            subs.append(f'{{{variable}}}')

            args, kwargs = [], {}

            if arguments:
                args, kwargs = parse_converter_args(arguments)

            schema = None
            if converter == 'any':
                schema = {
                    'type': 'array',
                    'items': {
                        'type': 'string',
                        'enum': args,
                    }
                }
            elif converter == 'int':
                schema = {
                    'type': 'integer',
                    'format': 'int32',
                }
                if 'max' in kwargs:
                    schema['maximum'] = kwargs['max']
                if 'min' in kwargs:
                    schema['minimum'] = kwargs['min']
            elif converter == 'float':
                schema = {
                    'type': 'number',
                    'format': 'float',
                }
            elif converter == 'uuid':
                schema = {
                    'type': 'string',
                    'format': 'uuid',
                }
            elif converter == 'path':
                schema = {
                    'type': 'string',
                    'format': 'path',
                }
            elif converter == 'string':
                schema = {
                    'type': 'string',
                }
                for prop in ['length', 'maxLength', 'minLength']:
                    if prop in kwargs:
                        schema[prop] = kwargs[prop]
            elif converter == 'default':
                schema = {'type': 'string'}

            parameters.append({
                'name': variable,
                'in': 'path',
                'required': True,
                'schema': schema,
            })

        return ''.join(subs), parameters

    def validate(self, *args, **kwargs):
        from flask import request, abort, make_response, jsonify

        func, query, json, headers, cookies, resp = pop_keywords(kwargs)

        try:
            arg = request.args or {}
            data = request.get_json() or {}
            head = request.headers or {}
            cook = request.cookies or {}
            request.context = Context(
                query(**arg) if query else None,
                json(**data) if json else None,
                headers(**head) if headers else None,
                cookies(**cook) if cookies else None,
            )
        except ValidationError as err:
            abort(make_response(jsonify(err.errors()), 422))
        except Exception:
            raise

        response = make_response(func(*args, **kwargs))

        if resp and resp.has_model():
            model = resp.find_model(response.status_code)
            model.validate(response.get_json())

        return response

    def register_route(self, app):
        self.app = app
        from flask import jsonify

        self.app.add_url_rule(
            self.config.spec_url,
            'openapi',
            lambda: jsonify(self.spectree.spec),
        )

        for ui in PAGES:
            self.app.add_url_rule(
                f'/{self.config.PATH}/{ui}',
                f'doc_page_{ui}',
                lambda ui=ui: PAGES[ui].format(self.config.spec_url)
            )



from collections import namedtuple

Context = namedtuple('Context', ['query', 'json', 'headers', 'cookies'])


class BasePlugin:
    """
    Base plugin for SpecTree plugin classes.

    :param spectree: :class:`spectree.SpecTree` instance
    """

    def __init__(self, spectree):
        self.spectree = spectree
        self.config = spectree.config

    def register_route(self, app):
        """
        :param app: backend framework application

        register document API routes to application
        """
        raise NotImplementedError

    def validate(self, query, json, headers, resp):
        """
        validate the request and response
        """
        raise NotImplementedError

    def find_routes(self):
        """
        find the routes from application
        """
        raise NotImplementedError

    def bypass(self, func, method):
        """
        :param func: route function (endpoint)
        :param method: HTTP method for this route function

        bypass some routes that shouldn't be shown in document
        """
        raise NotImplementedError

    def parse_path(self, route):
        """
        :param route: API routes

        parse URI path to get the variables in path
        """
        raise NotImplementedError

    def parse_func(self, route):
        """
        :param route: API routes

        get the endpoint function from routes
        """
        raise NotImplementedError



from functools import partial, wraps
from pydantic import BaseModel

from .plugins import PLUGINS
from .config import Config
from .utils import parse_comments, parse_request, parse_params, parse_resp, parse_name


class SpecTree:
    """
    Interface

    :param str backend: choose from ('flask', 'falcon', 'starlette')
    :param app: backend framework application instance (you can also register to it later)
    :param kwargs: update default :class:`spectree.config.Config`
    """

    def __init__(self, backend='base', app=None, **kwargs):
        self.config = Config(**kwargs)
        self.backend = PLUGINS[backend](self)
        # init
        self.models = {}
        if app:
            self.register(app)

    def register(self, app):
        """
        register to backend application

        This will be automatically triggered if the app is passed into the
        init step.
        """
        self.app = app
        self.backend.register_route(self.app)

    @property
    def spec(self):
        """
        get the OpenAPI spec
        """
        if not hasattr(self, '_spec'):
            self._spec = self._generate_spec()
        return self._spec

    def bypass(self, func):
        """
        bypass rules for routes (mode defined in config)

        :normal:    collect all the routes that are not decorated by other
                    `SpecTree` instance
        :greedy:    collect all the routes
        :strict:    collect all the routes decorated by this instance
        """
        if self.config.MODE == 'greedy':
            return False
        elif self.config.MODE == 'strict':
            if getattr(func, '_decorator', None) == self:
                return False
            return True
        else:
            decorator = getattr(func, '_decorator', None)
            if decorator and decorator != self:
                return True
            return False

    def validate(self, query=None, json=None, headers=None, cookies=None, resp=None, tags=[]):
        """
        - validate query, json, headers in request
        - validate response body and status code
        - add tags to this API route

        :param query: `pydantic.BaseModel`, query in uri like `?name=value`
        :param json: `pydantic.BaseModel`, JSON format request body
        :param headers: `pydantic.BaseModel`, if you have specific headers
        :param cookies: `pydantic.BaseModel`, if you have cookies for this route
        :param resp: `spectree.Response`
        :param tags: list of tags' string
        """
        def decorate_validation(func):
            validation = wraps(func)(partial(
                self.backend.validate,
                func=func, query=query, json=json, headers=headers,
                cookies=cookies, resp=resp))

            # register
            for name, model in zip(('query', 'json', 'headers', 'cookies'),
                                   (query, json, headers, cookies)):
                if model is not None:
                    assert(issubclass(model, BaseModel))
                    self.models[model.__name__] = model.schema()
                    setattr(validation, name, model.__name__)

            if resp:
                for model in resp.models:
                    self.models[model.__name__] = model.schema()
                validation.resp = resp

            if tags:
                validation.tags = tags

            # register decorator
            validation._decorator = self
            return validation
        return decorate_validation

    def _generate_spec(self):
        """
        generate OpenAPI spec according to routes and decorators
        """
        routes, tags = {}, {}
        for route in self.backend.find_routes():
            path, parameters = self.backend.parse_path(route)
            routes[path] = routes.get(path, {})
            for method, func in self.backend.parse_func(route):
                if self.backend.bypass(func, method) or self.bypass(func):
                    continue

                name = parse_name(func)
                summary, desc = parse_comments(func)
                func_tags = getattr(func, 'tags', [])
                for tag in func_tags:
                    if tag not in tags:
                        tags[tag] = {'name': tag}

                routes[path][method.lower()] = {
                    'summary': summary or f'{name} <{method}>',
                    'operationID': f'{name}__{method.lower()}',
                    'description': desc or '',
                    'tags': getattr(func, 'tags', []),
                    'requestBody': parse_request(func),
                    'parameters': parse_params(func, parameters[:]),
                    'responses': parse_resp(func),
                }

        spec = {
            'openapi': self.config.OPENAPI_VERSION,
            'info': {
                'title': self.config.TITLE,
                'version': self.config.VERSION,
            },
            'tags': list(tags.values()),
            'paths': {**routes},
            'components': {
                'schemas': {**self.models}
            },
            'definitions': self._get_model_definitions()
        }
        return spec

    def _get_model_definitions(self):
        """
        handle nested models
        """
        definitions = {}
        for schema in self.models.values():
            if 'definitions' in schema:
                for key, value in schema['definitions'].items():
                    definitions[key] = value
                del schema['definitions']

        return definitions



import re
import inspect

# parse HTTP status code to get the code
HTTP_CODE = re.compile(r'^HTTP_(?P<code>\d{3})$')


def parse_comments(func):
    """
    parse function comments

    First line of comments will be saved as summary, and the rest
    will be saved as description.
    """
    doc = inspect.getdoc(func)
    if doc is None:
        return None, None
    doc = doc.split('\n', 1)
    if len(doc) == 1:
        return doc[0], None
    return doc[0], doc[1].strip()


def parse_request(func):
    """
    get json spec
    """
    data = {
        'content': {
            'application/json': {
                'schema': {
                    '$ref': f'#/components/schemas/{func.json}'
                    if hasattr(func, 'json') else ''
                }
            }
        }
    }
    return data


def parse_params(func, params):
    """
    get spec for (query, headers, cookies)
    """
    if hasattr(func, 'query'):
        params.append({
            'name': func.query,
            'in': 'query',
            'required': True,
            'schema': {
                '$ref': f'#/components/schemas/{func.query}',
            }
        })
    if hasattr(func, 'headers'):
        params.append({
            'name': func.headers,
            'in': 'header',
            'schema': {
                '$ref': f'#/components/schemas/{func.headers}',
            }
        })
    if hasattr(func, 'cookies'):
        params.append({
            'name': func.cookies,
            'in': 'cookie',
            'schema': {
                '$ref': f'#/components/schemas/{func.cookies}',
            }
        })
    return params


def parse_resp(func):
    """
    get the response spec

    If this function does not have explicit ``resp`` but have other models,
    a ``422 Validation Error`` will be append to the response spec. Since
    this may be triggered in the validation step.
    """
    responses = {}
    if hasattr(func, 'resp'):
        responses = func.resp.generate_spec()

    if '422' not in responses and has_model(func):
        responses['422'] = {'description': 'Validation Error'}

    return responses


def has_model(func):
    """
    return True if this function have ``pydantic.BaseModel``
    """
    if any(hasattr(func, x) for x in ('query', 'json', 'headers')):
        return True

    if hasattr(func, 'resp') and func.resp.has_model():
        return True

    return False


def parse_code(http_code):
    """
    get the code of this HTTP status

    :param str http_code: format like ``HTTP_200``
    """
    match = HTTP_CODE.match(http_code)
    if not match:
        return None
    return match.group('code')


def parse_name(func):
    """
    the func can be

        * undecorated functions
        * decorated functions
        * decorated class methods
    """
    return func.__name__


def pop_keywords(kwargs):
    """
    pop (query, json, headers, cookies, resp, func) from the decorated function
    """
    query = kwargs.pop('query')
    json = kwargs.pop('json')
    headers = kwargs.pop('headers')
    cookies = kwargs.pop('cookies')
    resp = kwargs.pop('resp')
    func = kwargs.pop('func')
    return func, query, json, headers, cookies, resp



import re
import inspect
from functools import partial
from pydantic import ValidationError

from ..utils import pop_keywords
from .base import BasePlugin
from .page import PAGES


class OpenAPI:
    def __init__(self, spec):
        self.spec = spec

    def on_get(self, req, resp):
        resp.media = self.spec


class DocPage:
    def __init__(self, html, spec_url):
        self.page = html.format(spec_url)

    def on_get(self, req, resp):
        resp.content_type = 'text/html'
        resp.body = self.page


DOC_CLASS = [x.__name__ for x in (DocPage, OpenAPI)]


class FlaconPlugin(BasePlugin):
    def __init__(self, spectree):
        super().__init__(spectree)
        from falcon.routing.compiled import _FIELD_PATTERN

        self.FIELD_PATTERN = _FIELD_PATTERN
        # NOTE from `falcon.routing.compiled.CompiledRouterNode`
        self.ESCAPE = r'[\.\(\)\[\]\?\$\*\+\^\|]'
        self.ESCAPE_TO = r'\\\g<0>'
        self.EXTRACT = r'{\2}'
        # NOTE this regex is copied from werkzeug.routing._converter_args_re and
        # modified to support only int args
        self.INT_ARGS = re.compile(r'''
            ((?P<name>\w+)\s*=\s*)?
            (?P<value>\d+)\s*
        ''', re.VERBOSE)
        self.INT_ARGS_NAMES = ('num_digits', 'min', 'max')

    def register_route(self, app):
        self.app = app
        self.app.add_route(
            self.config.spec_url, OpenAPI(self.spectree.spec)
        )
        for ui in PAGES:
            self.app.add_route(
                f'/{self.config.PATH}/{ui}',
                DocPage(PAGES[ui], self.config.spec_url),
            )

    def find_routes(self):
        routes = []

        def find_node(node):
            if node.resource and node.resource.__class__.__name__ not in DOC_CLASS:
                routes.append(node)

            for child in node.children:
                find_node(child)

        for route in self.app._router._roots:
            find_node(route)

        return routes

    def parse_func(self, route):
        return route.method_map.items()

    def parse_path(self, route):
        subs, parameters = [], []
        for segment in route.uri_template.strip('/').split('/'):
            matches = self.FIELD_PATTERN.finditer(segment)
            if not matches:
                subs.append(segment)
                continue

            escaped = re.sub(self.ESCAPE, self.ESCAPE_TO, segment)
            subs.append(self.FIELD_PATTERN.sub(self.EXTRACT, escaped))

            for field in matches:
                variable, converter, argstr = [field.group(name) for name in
                                               ('fname', 'cname', 'argstr')]

                if converter == 'int':
                    if argstr is None:
                        argstr = ''

                    arg_values = [None, None, None]
                    for index, match in enumerate(self.INT_ARGS.finditer(argstr)):
                        name, value = match.group('name'), match.group('value')
                        if name:
                            index = self.INT_ARGS_NAMES.index(name)
                        arg_values[index] = value

                    num_digits, minumum, maximum = arg_values
                    schema = {
                        'type': 'integer',
                        'format': f'int{num_digits}' if num_digits else 'int32',
                    }
                    if minumum:
                        schema['minimum'] = minumum
                    if maximum:
                        schema['maximum'] = maximum
                elif converter == 'uuid':
                    schema = {
                        'type': 'string',
                        'format': 'uuid'
                    }
                elif converter == 'dt':
                    schema = {
                        'type': 'string',
                        'format': 'date-time',
                    }
                else:
                    # no converter specified or customized converters
                    schema = {'type': 'string'}

                parameters.append({
                    'name': variable,
                    'in': 'path',
                    'required': True,
                    'schema': schema,
                })

        return f'/{"/".join(subs)}', parameters

    def validate(self, _req, _resp, *args, **kwargs):
        func, query, json, headers, cookies, resp = pop_keywords(kwargs)
        try:
            if query:
                _req.context.query = query(**_req.params)
            if headers:
                _req.context.headers = headers(**_req.headers)
            if cookies:
                _req.context.cookies = cookies(**_req.cookies)
            media = _req.media or {}
            if json:
                _req.context.json = json(**media)

        except ValidationError as err:
            _resp.status = '422 Unprocessable Entity'
            _resp.media = err.errors()
            return
        except Exception:
            raise

        func(self, _req, _resp, *args, **kwargs)
        if resp and resp.has_model():
            model = resp.find_model(_resp.status[:3])
            model.validate(_resp.media)

    def bypass(self, func, method):
        if not isinstance(func, partial):
            return False
        if inspect.ismethod(func.func):
            return False
        # others are <cyfunction>
        return True



from pydantic import BaseModel

from .utils import parse_code


class Response:
    """
    response object

    :param codes: list of HTTP status code, format('HTTP_[0-9]{3}'), 'HTTP200'
    :param code_models: dict of <HTTP status code>: <`pydantic.BaseModel`>
    """

    def __init__(self, *codes, **code_models):
        for code in codes:
            assert code in DEFAULT_CODE_DESC, 'invalid HTTP status code'

        for code, model in code_models.items():
            assert code in DEFAULT_CODE_DESC, 'invalid HTTP status code'
            assert issubclass(model, BaseModel), 'invalid `pydantic.BaseModel`'

        self.codes = codes
        self.code_models = code_models

    def has_model(self):
        """
        :returns: boolean -- does this response has models or not
        """
        return True if self.code_models else False

    def find_model(self, code):
        """
        :param code: ``r'\\d{3}'``
        """
        return self.code_models.get(f'HTTP_{code}')

    @property
    def models(self):
        """
        :returns:  dict_values -- all the models in this response
        """
        return self.code_models.values()

    def generate_spec(self):
        """
        generate the spec for responses

        :returns: JSON
        """
        responses = {}
        for code in self.codes:
            responses[parse_code(code)] = {'description': DEFAULT_CODE_DESC[code]}

        for code, model in self.code_models.items():
            responses[parse_code(code)] = {
                'description': DEFAULT_CODE_DESC[code],
                'content': {
                    'application/json': {
                        'schema': {
                            '$ref': f'#/components/schemas/{model.__name__}'
                        }
                    }
                }
            }

        return responses


# according to https://tools.ietf.org/html/rfc2616#section-10
# https://tools.ietf.org/html/rfc7231#section-6.1
# https://developer.mozilla.org/sv-SE/docs/Web/HTTP/Status
DEFAULT_CODE_DESC = {
    # Information 1xx
    'HTTP_100': 'Continue',
    'HTTP_101': 'Switching Protocols',
    # Successful 2xx
    'HTTP_200': 'OK',
    'HTTP_201': 'Created',
    'HTTP_202': 'Accepted',
    'HTTP_203': 'Non-Authoritative Information',
    'HTTP_204': 'No Content',
    'HTTP_205': 'Reset Content',
    'HTTP_206': 'Partial Content',
    # Redirection 3xx
    'HTTP_300': 'Multiple Choices',
    'HTTP_301': 'Moved Permanently',
    'HTTP_302': 'Found',
    'HTTP_303': 'See Other',
    'HTTP_304': 'Not Modified',
    'HTTP_305': 'Use Proxy',
    'HTTP_306': '(Unused)',
    'HTTP_307': 'Temporary Redirect',
    'HTTP_308': 'Permanent Redirect',
    # Client Error 4xx
    'HTTP_400': 'Bad Request',
    'HTTP_401': 'Unauthorized',
    'HTTP_402': 'Payment Required',
    'HTTP_403': 'Forbidden',
    'HTTP_404': 'Not Found',
    'HTTP_405': 'Method Not Allowed',
    'HTTP_406': 'Not Acceptable',
    'HTTP_407': 'Proxy Authentication Required',
    'HTTP_408': 'Request Timeout',
    'HTTP_409': 'Conflict',
    'HTTP_410': 'Gone',
    'HTTP_411': 'Length Required',
    'HTTP_412': 'Precondition Failed',
    'HTTP_413': 'Request Entity Too Large',
    'HTTP_414': 'Request-URI Too Long',
    'HTTP_415': 'Unsupported Media Type',
    'HTTP_416': 'Requested Range Not Satisfiable',
    'HTTP_417': 'Expectation Failed',
    'HTTP_418': "I'm a teapot",
    'HTTP_421': 'Misdirected Request',
    'HTTP_422': 'Unprocessable Entity',
    'HTTP_423': 'Locked',
    'HTTP_424': 'Failed Dependency',
    'HTTP_425': 'Too Early',
    'HTTP_426': 'Upgrade Required',
    'HTTP_428': 'Precondition Required',
    'HTTP_429': 'Too Many Requests',
    'HTTP_431': 'Request Header Fields Too Large',
    'HTTP_451': 'Unavailable For Legal Reasons',
    # Server Error 5xx
    'HTTP_500': 'Internal Server Error',
    'HTTP_501': 'Not Implemented',
    'HTTP_502': 'Bad Gateway',
    'HTTP_503': 'Service Unavailable',
    'HTTP_504': 'Gateway Timeout',
    'HTTP_505': 'HTTP Version Not Supported',
    'HTTP_506': 'Variant Also negotiates',
    'HTTP_507': 'Insufficient Sotrage',
    'HTTP_508': 'Loop Detected',
    'HTTP_511': 'Network Authentication Required',
}



class Config:
    """
    :ivar MODE: mode for route. **normal** includes undecorated routes and
        routes decorated by this instance. **strict** only includes routes
        decorated by this instance. **greedy** includes all the routes.
    :ivar PATH: path for API document page
    :ivar UI: UI for API document, 'redoc' or 'swagger'
    :ivar OPENAPI_VERSION: OpenAPI version
    :ivar TITLE: service name
    :ivar VERSION: service version
    :ivar DOMAIN: service host domain
    """

    def __init__(self, **kwargs):
        self.PATH = 'apidoc'
        self.FILENAME = 'openapi.json'
        self.OPENAPI_VERSION = '3.0.2'
        self.UI = 'redoc'
        self._SUPPORT_UI = {'redoc', 'swagger'}
        self.MODE = 'normal'
        self._SUPPORT_MODE = {'normal', 'strict', 'greedy'}

        self.TITLE = 'Service API Document'
        self.VERSION = '0.1'
        self.DOMAIN = None

        self.update(**kwargs)

    @property
    def spec_url(self):
        return f'/{self.PATH}/{self.FILENAME}'

    def __repr__(self):
        display = '\n{:=^80}\n'.format(self.__class__.__name__)
        for k, v in vars(self).items():
            if not k.startswith('__'):
                display += '| {:<30} {}\n'.format(k, v)

        return display + '=' * 80

    def update(self, **kwargs):
        """
        update config from key-value pairs

        :param kwargs: key(case insensitive)-value pairs for config
        """
        for key, value in kwargs.items():
            key = key.upper()
            if not hasattr(self, key):
                print(f'[✗] Ignore unknown attribute "{key}"')
            else:
                setattr(self, key, value)
                print(f'[✓] Attribute "{key}" has been updated to "{value}"')

        assert self.UI in self._SUPPORT_UI, 'unsupported UI'
        assert self.MODE in self._SUPPORT_MODE, 'unsupported MODE'



import inspect
from json import loads as json_loads
from collections import namedtuple
from functools import partial
from pydantic import ValidationError

from ..utils import pop_keywords
from .base import BasePlugin, Context
from .page import PAGES


METHODS = {'get', 'post', 'put', 'patch', 'delete'}
Route = namedtuple('Route', ['path', 'methods', 'func'])


class StarlettePlugin(BasePlugin):
    def __init__(self, spectree):
        super().__init__(spectree)
        from starlette.convertors import CONVERTOR_TYPES
        self.conv2type = {
            conv: typ for typ, conv in CONVERTOR_TYPES.items()
        }

    def register_route(self, app):
        self.app = app
        from starlette.responses import JSONResponse, HTMLResponse

        self.app.add_route(
            self.config.spec_url,
            lambda request: JSONResponse(self.spectree.spec),
        )

        for ui in PAGES:
            self.app.add_route(
                f'/{self.config.PATH}/{ui}',
                lambda request, ui=ui: HTMLResponse(
                    PAGES[ui].format(self.config.spec_url)
                ),
            )

    async def validate(self, scope, receive, send, *args, **kwargs):
        from starlette.requests import Request
        from starlette.responses import JSONResponse

        func, query, json, headers, cookies, resp = pop_keywords(kwargs)

        request = Request(scope, receive)
        try:
            request.context = Context(
                query(**request.query_params) if query else None,
                json(**json_loads(await request.body() or '{}')) if json else None,
                headers(**request.headers) if headers else None,
                cookies(**request.cookies) if cookies else None,
            )
        except ValidationError as err:
            response = JSONResponse(err.errors(), 422)
            await response(scope, receive, send)
            return
        except Exception:
            raise

        if inspect.iscoroutinefunction(func):
            response = await func(request, *args, **kwargs)
        else:
            response = func(request, *args, **kwargs)

        if resp:
            model = resp.find_model(response.status_code)
            model.validate(json_loads(response.body))

        await response(scope, receive, send)

    def find_routes(self):
        routes = []

        def parse_route(app, prefix=''):
            for route in app.routes:
                if route.path.startswith(f'/{self.config.PATH}'):
                    continue

                func = route.endpoint
                if isinstance(func, partial):
                    try:
                        func = func.__wrapped__
                    except AttributeError:
                        pass

                if inspect.isclass(func):
                    for method in METHODS:
                        if getattr(func, method, None):
                            routes.append(Route(
                                f'{prefix}{route.path}',
                                {method.upper()},
                                getattr(func, method)
                            ))
                elif inspect.isfunction(func):
                    routes.append(Route(
                        f'{prefix}{route.path}',
                        route.methods,
                        route.endpoint))
                else:
                    parse_route(route, prefix=f'{prefix}{route.path}')

        parse_route(self.app)
        return routes

    def bypass(self, func, method):
        if method in ['HEAD', 'OPTIONS']:
            return True
        return False

    def parse_func(self, route):
        for method in route.methods or ['GET']:
            yield method, route.func

    def parse_path(self, route):
        from starlette.routing import compile_path
        variables = compile_path(route.path)[-1]
        parameters = []

        for name, conv in variables.items():
            schema = None
            typ = self.conv2type[conv]
            if typ == 'int':
                schema = {
                    'type': 'integer',
                    'format': 'int32'
                }
            elif typ == 'float':
                schema = {
                    'type': 'number',
                    'format': 'float',
                }
            elif typ == 'path':
                schema = {
                    'type': 'string',
                    'format': 'path',
                }
            elif typ == 'str':
                schema = {'type': 'string'}

            parameters.append({
                'name': name,
                'in': 'path',
                'required': True,
                'schema': schema,
            })

        return route.path, parameters



from .base import BasePlugin
from .flask_plugin import FlaskPlugin
from .falcon_plugin import FlaconPlugin
from .starlette_plugin import StarlettePlugin

PLUGINS = {
    'base': BasePlugin,
    'flask': FlaskPlugin,
    'falcon': FlaconPlugin,
    'starlette': StarlettePlugin,
}
